import {AdEntity, NewAdEntity, SimpleAdEntity} from "../types";
import {ValidationError} from "../utlis/errors";
import {pool} from "../utlis/db";
import {FieldPacket} from "mysql2";
import {v4 as uuid} from "uuid";

type AdRecordResults = [AdEntity[], FieldPacket[]];

export class AdRecord implements AdEntity{
  id: string;
  name: string;
  desc: string;
  price: number;
  url: string;
  lat: number;
  lon: number;
  constructor(obj: NewAdEntity) {
    if(!obj.name || obj.name.length > 100){
      throw new ValidationError('Ad name cannot be empty and cannot exceed 100 characters');
    }

    if(obj.desc.length > 1000){
      throw new ValidationError('Ad description cannot exceed 1000 characters');
    }

    if(obj.price < 0 || obj.price > 9999999){
      throw new ValidationError('Price cannot be negative or greater than 9 999 999');
    }

    //@TODO: check if url is valid!
    if(!obj.url || obj.url.length > 100){
      throw new ValidationError('Ad url cannot be empty and cannot exceed 100 characters');
    }

    if(typeof obj?.lat !== 'number' || typeof obj?.lon !== 'number'){
      throw new ValidationError('Cannot locate ad');
    }

    obj.id ? this.id = obj.id : null;
    this.name = obj.name;
    this.desc = obj.desc;
    this.price = obj.price;
    this.url = obj.url;
    this.lat = obj.lat;
    this.lon = obj.lon;

  }
  static async getOne(id: string): Promise<AdRecord | null> {
    const [results] = await pool.execute("SELECT * FROM `ads` WHERE `id` = :id", {
      id,
    }) as AdRecordResults;

    return results.length === 0 ? null : new AdRecord(results[0]);
  }

  async insert(): Promise<string> {
    if(!this.id){
      this.id = uuid();
    } else {
      throw new Error('Cannot insert something that is already in database');
    }

    await pool.execute("INSERT INTO `ads` VALUES (:id, :name, :desc, :price, :url, :lat, :lon)", this);

    return this.id;
  }

  static async find(name: string): Promise<SimpleAdEntity[]> {
    const [results] = await pool.execute("SELECT * FROM `ads` WHERE `name` LIKE :search",{
      search: `%${name}%`,
    }) as AdRecordResults;

    return results.map(result => {
      const {id, lat, lon,} = result;
      return {
        id, lat, lon,
      }
    });
  }

}
